/*
 * Convert a jpeg on stdin to an interlaced C64 multicolour image on stdout
 *
 * Version 1.0.1
 *
 */

#include <stdio.h>
#include <jpeglib.h>
#ifdef DOS
#include <fcntl.h>
#endif

/* mcicode[1000] has the code */
#include "mcicode.h"


/* where the different parts go in memory */
#define A_CODE     0x0801
#define A_COLRAM   0x1000
#define A_CHARRAM0 0x0c00
#define A_CHARRAM1 0x1400
#define A_BITMAP0  0x2000
#define A_BITMAP1  0x4000
#define A_START 0x0801
#define A_END   0x5f40


/*UINT8 r[16];*/
UINT8 r[16] = {0x00, 0xff, 0x68, 0x70, 0x6f, 0x58, 0x35, 0xb8,
	       0x6f, 0x43, 0x9a, 0x44, 0x6c, 0x9a, 0x6c, 0x95};
UINT8 g[16] = {0x00, 0xff, 0x37, 0xa4, 0x3d, 0x8d, 0x28, 0xc7,
	       0x4f, 0x39, 0x67, 0x44, 0x6c, 0xd2, 0x5e, 0x95};
UINT8 b[16] = {0x00, 0xff, 0x2b, 0xb2, 0x86, 0x43, 0x79, 0x6f,
	       0x25, 0x00, 0x59, 0x44, 0x6c, 0x84, 0xb5, 0x95};


/* struct to hold an 8x8 pixel area of an mci pic */
typedef struct {
  unsigned char charram0;
  unsigned char charram1;
  unsigned char colram;
  unsigned char bitmap0[8];
  unsigned char bitmap1[8];
} mcichar;

unsigned char bkgcol;
unsigned char lastcol0, lastcol1, lastcol2;

int quiet, verbose;


long dist(int r1, int g1, int b1, int r2, int g2, int b2) {
  return((r1-r2)*(r1-r2) + (g1-g2)*(g1-g2) + (b1-b2)*(b1-b2));
}


long vicdist(int c1, int c2) {
  return((r[c1]-r[c2])*(r[c1]-r[c2]) + (g[c1]-g[c2])*(g[c1]-g[c2]) + (b[c1]-b[c2])*(b[c1]-b[c2]));
}


UINT8 viccol(int r1, int g1, int b1) {
  UINT8 c;
  long d;
  int i;

  //printf("mapping %3d %3d %3d\r\n", r1, g1, b1);
  d = dist(r1, g1, b1, r[0], g[0], b[0]);
  c = 0;
  //printf("dist  0: %d\r\n", d);
  for (i = 0; i < 16; i++) {
    if (dist(r1, g1, b1, r[i], g[i], b[i]) < d) {
      d = dist(r1, g1, b1, r[i], g[i], b[i]);
      c = i;
      //printf("dist %2d: %d\r\n", i, d);
    }
  }
  return(c);
}


/* pick the best match out of the four provided */
int pickonecolour(int colour, int choice0, int choice1, int choice2, int choice3)
{
  int dist, mindist, result;

  if (colour == choice0) {
    return(0);
  } else if (colour == choice1) {
    return(1);
  } else if (colour == choice2) {
    return(2);
  } else if (colour == choice3) {
    return(3);
  } else {
    /* now we're getting ugly */
    result = 0;
    mindist = vicdist(colour, choice0);
    if ((dist = vicdist(colour, choice1)) < mindist) {
      result = 1;
      mindist = dist;
    }
    if ((dist = vicdist(colour, choice2)) < mindist) {
      result = 2;
      mindist = dist;
    }
    if ((dist = vicdist(colour, choice3)) < mindist) {
      return(3);
    }
    return(result);
  }
}


/* find the optimal background colour */
UINT8 findbkg(UINT8 *pic) {
  int row, col, x, y;
  unsigned int i[16], c[16];
  int a;

  for (x = 0; x < 16; x++) {
    c[x] = 0;
  }
  for (row = 0; row < 200; row += 8) {
    for (col = 0; col < 320; col += 8) {
      for (x = 0; x < 16; x++) {
	i[x] = 0;
      }
      for (y = 0; y < 8; y++) {
	for (x = 0; x < 8; x++) {
	  i[pic[(row+y)*320+col+x]] = 1;
	}
      }
      for (x = 0; x < 16; x++) {
	c[x] += i[x];
      }
    }
  }
  a = 0;
  for (x = 0; x < 16; x++) {
    //printf("%2d:%4d\r\n", x, c[x]);
    if (c[x] > c[a]) {
      a = x;
    }
  }
  return(a);
}


/* convert an 8x8 buffer of c64 colours to mci (two 8x8 bitmaps, two charrams,
   shared colram */
void convertmcichar(UINT8 *source, /* UINT8 *dest, */ mcichar *mc)
{
  int colcount[16];
  int sortcols[16];
  int col[6];
  int i, j;
  int key, colkey, choice;

  for (i = 0; i < 16; i++) {
    sortcols[i] = i;
    colcount[i] = 0;
  }

  /* count colours */
  for (i = 0; i < 64; i++)
    colcount[source[i]] += 1;
  colcount[bkgcol] = 0;
  col[3] = 0;
  for (i = 1; i < 16; i++) {
    if (colcount[i] > col[3]) {
      col[3] = i;
    }
  }

  for (i = 0; i < 16; i++) {
    sortcols[i] = i;
    colcount[i] = 0;
  }

  /* count colours */
  for (i = 0; i < 64; i += 2)
    colcount[source[i]] += 1;

  /* don't want the background colour to interfere with colour selection */
  colcount[bkgcol] = 0;
  colcount[col[3]] = 0;

  /* sort by frequency */
  for (j = 1; j < 16; j++) {
    key = colcount[j];
    colkey = sortcols[j];
    i = j - 1;
    while (i >= 0 && colcount[i] < key) {
      colcount[i + 1] = colcount[i];
      sortcols[i + 1] = sortcols[i];
      i--;
    }
    colcount[i + 1] = key;
    sortcols[i + 1] = colkey;
  }

  /* select four colours */
  col[0] = bkgcol;      /* $d021 */
  col[1] = sortcols[0]; /* msn charram */
  col[2] = sortcols[1]; /* lsn charram */

  /* store colour selections */
  mc->colram = col[3];
  mc->charram0 = col[1]<<4 | col[2];

  for (i = 0; i < 16; i++) {
    sortcols[i] = i;
    colcount[i] = 0;
  }
  for (i = 1; i < 64; i += 2)
    colcount[source[i]] += 1;
  colcount[bkgcol] = 0;
  colcount[col[3]] = 0;
  for (j = 1; j < 16; j++) {
    key = colcount[j];
    colkey = sortcols[j];
    i = j - 1;
    while (i >= 0 && colcount[i] < key) {
      colcount[i + 1] = colcount[i];
      sortcols[i + 1] = sortcols[i];
      i--;
    }
    colcount[i + 1] = key;
    sortcols[i + 1] = colkey;
  }
  col[4] = sortcols[0];
  col[5] = sortcols[1];

  mc->charram1 = col[4]<<4 | col[5];

  /* save selections for next conversion */
  lastcol0 = col[1];
  lastcol1 = col[2];
  lastcol2 = col[3];

  {
    int x, y;
    UINT8 pixel[8];

    /* render using the new colours */
    for (y = 0; y < 8; y++) {
      for (x = 0; x < 8; x++) {
	if (x & 1) {
	  pixel[x] = pickonecolour(source[y*8+x], col[0], col[4], col[5], col[3]);
	} else {
	  pixel[x] = pickonecolour(source[y*8+x], col[0], col[1], col[2], col[3]);
	}
      }
      mc->bitmap0[y] = (pixel[0] & 3)<<6 | (pixel[2] & 3)<<4 | (pixel[4] & 3)<<2 | (pixel[6] & 3);
      mc->bitmap1[y] = (pixel[1] & 3)<<6 | (pixel[3] & 3)<<4 | (pixel[5] & 3)<<2 | (pixel[7] & 3);
    }
  }
}


void storeline(JSAMPLE *line, int row_stride, UINT8 *rawimage, int linenum) {
  int i, x;
  int offset;


  offset = linenum * row_stride;
  for (i = 0; i < row_stride; i += 3) {
    /* this won't work if jpeglib doesn't use 8 bits per component */
    rawimage[offset + i  ] = line[i  ];
    rawimage[offset + i+1] = line[i+1];
    rawimage[offset + i+2] = line[i+2];
  }
}


#define RangeLH(x,min,max) {if(x>max)x=max;else if(x<min)x=min;}

/* reduce 320x200x3 to 320x200x1 with vic palette */
void reduce(UINT8 *source, UINT8 *dest) {
  int xx,yy,thiscolor;

  int sR,sG,sB,r1,r2,r3,r4,g1,g2,g3,g4,b1,b2,b3,b4;

  UINT8 work[3][320][200];  // temp. work area

  float dR1,dR2,dR3,dR4,eR;
  float dG1,dG2,dG3,dG4,eG;
  float dB1,dB2,dB3,dB4,eB;

  float ds = 1.0;

  int pO_x,pO_y,pSO_x,pSO_y,pS_x,pS_y,pSW_x,pSW_y;    // pixel coordinates

  UINT8 *srcline;
  UINT8 *destline;

  //-- make a working copy of source pic --//
 
  for (yy = 0; yy < 200; yy++) {
    srcline = source + yy*320*3; // 3 bytes per pixel
    for (xx = 0; xx < 320; xx++) {
      work[0][xx][yy] = srcline[xx*3  ];
      work[1][xx][yy] = srcline[xx*3+1];
      work[2][xx][yy] = srcline[xx*3+2];
    }
  }

  //-- color reduction --//
  for (yy = 0; yy < 200; yy++) {
    destline = dest + yy*320; // 1 byte per pixel
    for (xx = 0; xx < 320; xx++) {

      // get source cols
      sR = work[0][xx][yy];  sG = work[1][xx][yy];  sB = work[2][xx][yy];

      // approximate destination luma
      thiscolor = viccol(sR, sG, sB);

      // calc error
      eR = ds * (sR - r[thiscolor]);
      eG = ds * (sG - g[thiscolor]);
      eB = ds * (sB - b[thiscolor]);

      // set destination pixel
      destline[xx] = thiscolor;

      pO_x  = xx+1; pO_y  = yy;
      pSO_x = xx+1; pSO_y = yy+1;
      pS_x  = xx;   pS_y  = yy+1;
      pSW_x = xx-1; pSW_y = yy+1;

      RangeLH(pO_x,0,319);   RangeLH(pO_y,0,199);
      RangeLH(pSO_x,0,319);  RangeLH(pSO_y,0,199);
      RangeLH(pS_x,0,319);   RangeLH(pS_y,0,199);
      RangeLH(pSW_x,0,319);  RangeLH(pSW_y,0,199);

      dR1 = ((7*eR)/16);  dR2 = ((3*eR)/16);
      dR3 = ((5*eR)/16);  dR4 = eR-(dR1+dR2+dR3);
      dG1 = ((7*eG)/16);  dG2 = ((3*eG)/16);
      dG3 = ((5*eG)/16);  dG4 = eG-(dG1+dG2+dG3);
      dB1 = ((7*eB)/16);  dB2 = ((3*eB)/16);
      dB3 = ((5*eB)/16);  dB4 = eB-(dB1+dB2+dB3);

      r1 = work[0][pO_x][pO_y];   r2 = work[0][pSW_x][pSW_y];
      r3 = work[0][pS_x][pS_y];   r4 = work[0][pSO_x][pSO_y];
      g1 = work[1][pO_x][pO_y];   g2 = work[1][pSW_x][pSW_y];
      g3 = work[1][pS_x][pS_y];   g4 = work[1][pSO_x][pSO_y];
      b1 = work[2][pO_x][pO_y];   b2 = work[2][pSW_x][pSW_y];
      b3 = work[2][pS_x][pS_y];   b4 = work[2][pSO_x][pSO_y];

      r1 += dR1;  r2 += dR2;  r3 += dR3;  r4 += dR4;
      g1 += dG1;  g2 += dG2;  g3 += dG3;  g4 += dG4;
      b1 += dB1;  b2 += dB2;  b3 += dB3;  b4 += dB4;

      RangeLH(r1,0,255);  RangeLH(r2,0,255);  RangeLH(r3,0,255);  RangeLH(r4,0,255);
      RangeLH(g1,0,255);  RangeLH(g2,0,255);  RangeLH(g3,0,255);  RangeLH(g4,0,255);
      RangeLH(b1,0,255);  RangeLH(b2,0,255);  RangeLH(b3,0,255);  RangeLH(b4,0,255);

      work[0][pO_x][pO_y] = r1;  work[0][pSW_x][pSW_y] = r2;
      work[0][pS_x][pS_y] = r3;  work[0][pSO_x][pSO_y] = r4;
      work[1][pO_x][pO_y] = g1;  work[1][pSW_x][pSW_y] = g2;
      work[1][pS_x][pS_y] = g3;  work[1][pSO_x][pSO_y] = g4;
      work[2][pO_x][pO_y] = b1;  work[2][pSW_x][pSW_y] = b2;
      work[2][pS_x][pS_y] = b3;  work[2][pSO_x][pSO_y] = b4;
    }

    yy++;
    destline = dest + yy*320; // 1 byte per pixel

    for(xx = 319; xx > -1; xx--) {

      // get source cols
      sR = work[0][xx][yy];  sG = work[1][xx][yy];  sB = work[2][xx][yy];

      // approximate destination luma
      thiscolor = viccol(sR, sG, sB);

      // calc error
      eR = sR-r[thiscolor];
      eG = sG-g[thiscolor];
      eB = sB-b[thiscolor];

      // set destination pixel
      destline[xx] = thiscolor;

      pO_x = xx-1; pO_y = yy;
      pSO_x = xx-1; pSO_y = yy+1;
      pS_x = xx; pS_y = yy+1;
      pSW_x = xx+1; pSW_y = yy+1;

      RangeLH(pO_x,0,319);   RangeLH(pO_y,0,199);
      RangeLH(pSO_x,0,319);  RangeLH(pSO_y,0,199);
      RangeLH(pS_x,0,319);   RangeLH(pS_y,0,199);
      RangeLH(pSW_x,0,319);  RangeLH(pSW_y,0,199);

      dR1 = ((7*eR)/16);  dR2 = ((3*eR)/16);
      dR3 = ((5*eR)/16);  dR4 = eR-(dR1+dR2+dR3);
      dG1 = ((7*eG)/16);  dG2 = ((3*eG)/16);
      dG3 = ((5*eG)/16);  dG4 = eG-(dG1+dG2+dG3);
      dB1 = ((7*eB)/16);  dB2 = ((3*eB)/16);
      dB3 = ((5*eB)/16);  dB4 = eB-(dB1+dB2+dB3);

      r1 = work[0][pO_x][pO_y];   r2 = work[0][pSW_x][pSW_y];
      r3 = work[0][pS_x][pS_y];   r4 = work[0][pSO_x][pSO_y];
      g1 = work[1][pO_x][pO_y];   g2 = work[1][pSW_x][pSW_y];
      g3 = work[1][pS_x][pS_y];   g4 = work[1][pSO_x][pSO_y];
      b1 = work[2][pO_x][pO_y];   b2 = work[2][pSW_x][pSW_y];
      b3 = work[2][pS_x][pS_y];   b4 = work[2][pSO_x][pSO_y];

      r1 += dR1;  r2 += dR2;  r3 += dR3;  r4 += dR4;
      g1 += dG1;  g2 += dG2;  g3 += dG3;  g4 += dG4;
      b1 += dB1;  b2 += dB2;  b3 += dB3;  b4 += dB4;

      RangeLH(r1,0,255);  RangeLH(r2,0,255);  RangeLH(r3,0,255);  RangeLH(r4,0,255);
      RangeLH(g1,0,255);  RangeLH(g2,0,255);  RangeLH(g3,0,255);  RangeLH(g4,0,255);
      RangeLH(b1,0,255);  RangeLH(b2,0,255);  RangeLH(b3,0,255);  RangeLH(b4,0,255);

      work[0][pO_x][pO_y] = r1;  work[0][pSW_x][pSW_y] = r2;
      work[0][pS_x][pS_y] = r3;  work[0][pSO_x][pSO_y] = r4;
      work[1][pO_x][pO_y] = g1;  work[1][pSW_x][pSW_y] = g2;
      work[1][pS_x][pS_y] = g3;  work[1][pSO_x][pSO_y] = g4;
      work[2][pO_x][pO_y] = b1;  work[2][pSW_x][pSW_y] = b2;
      work[2][pS_x][pS_y] = b3;  work[2][pSO_x][pSO_y] = b4;
    }
  }
}


int main(int argc, char **argv) {
  struct jpeg_decompress_struct cinfo;
  struct jpeg_error_mgr jerr;
  JSAMPARRAY buffer;   /* Output row buffer */
  int row_stride;      /* physical row width in output buffer */
  UINT8 *rawimage;     /* will contain the decoded image */
  int width, height;
  float s, sx, sy;
  int x, y, xoff, ypos;
  UINT8 pic[320*200];  /* 16-colour pic here */
  UINT8 *line;
  UINT8 scaledpic[320*200*3]; /* 320x200 pic */
  UINT8 *scaledline;
  UINT8 bkgcol;

  /* parse args */
  verbose = 0;
  quiet = 0;
  if (argc == 2) {
    if (strcmp(argv[1], "-v") == 0) {
      verbose = 1;
    } else if (strcmp(argv[1], "-q") == 0) {
      quiet = 1;
    } else {
      fprintf(stderr, "Usage: jpg2mci [-vq]\r\n");
      exit(1);
    }
  } else if (argc != 1) {
    fprintf(stderr, "Usage: jpg2mci [-vq]\r\n");
    exit(1);
  }

#ifdef DOS
  /* dos sucks */
  setmode(0, O_BINARY);
  setmode(1, O_BINARY);
#endif

  /* create jpeg object */
  cinfo.err = jpeg_std_error(&jerr);
  jpeg_create_decompress(&cinfo);

  /* assign stdin to jpeg object */
  jpeg_stdio_src(&cinfo, stdin);

  /* read the jpeg header */
  jpeg_read_header(&cinfo, TRUE);

  /* allocate the image buffer */
  width = cinfo.image_width;
  height = cinfo.image_height;
  if ((rawimage = (UINT8 *) malloc((size_t) (width * height * 3))) == NULL) {
    if (!quiet) fprintf(stderr, "can't allocate %d bytes of memory\r\n", width * height * 3);
    exit(1);
  }

  /* print dimensions */
  if (verbose) fprintf(stderr, "%dx%dx%d\r\n", width, height, cinfo.num_components*8);

  /* force RGB */
  cinfo.out_color_space = JCS_RGB;

  /* start decompression */
  jpeg_start_decompress(&cinfo);

  /* JSAMPLEs per row in output buffer */
  row_stride = cinfo.output_width * cinfo.output_components;
  if (row_stride != width * 3) {
    if (!quiet) fprintf(stderr, "row_stride != width * 3, bailing out\r\n");
    exit(1);
  }
  /* Make a one-row-high sample array that will go away when done with image */
  buffer = (*cinfo.mem->alloc_sarray) ((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);

  /* for each scanline ... */
  while (cinfo.output_scanline < cinfo.output_height) {
    (void) jpeg_read_scanlines(&cinfo, buffer, 1);
    storeline(buffer[0], row_stride, rawimage, cinfo.output_scanline-1);
  }

  /* cleanup */
  jpeg_finish_decompress(&cinfo);
  jpeg_destroy_decompress(&cinfo);

  /* figure out the scale of the image */
  s = 1;
  if (width > 320) {
    s = (float) width/320.0;
  }
  if (height > 200) {
    if ((float) height/200.0 > s) {
      s = (float) height/200.0;
    }
  }
  if (s != 1.0) {
    if (verbose) fprintf(stderr, "Scaling to %d%%\r\n", (int) (1.0 / s * 100.0));
  }

  /* fix this */
  bkgcol = 0;

  /* copy (and scale and pad if necessary) to 320x200 bitmap */
  for (y = 0; y < 200; y++) {
    ypos = y*320*3;
    if ((float) y * s < (float) height) {
      line = rawimage + row_stride * (int) (y * s);
      for (x = 0; x < 320; x++) {
	if ((float) x * s < (float) width) {
	  xoff = 3 * (int) (x * s);
	  scaledpic[ypos+x*3  ] = line[xoff];
	  scaledpic[ypos+x*3+1] = line[xoff+1];
	  scaledpic[ypos+x*3+2] = line[xoff+2];
	} else {
	  scaledpic[ypos+x*3  ] = 0;
	  scaledpic[ypos+x*3+1] = 0;
	  scaledpic[ypos+x*3+2] = 0;
	}
      }
    } else {
      for (x = 0; x < 320; x++) {
	scaledpic[ypos+x*3  ] = 0;
	scaledpic[ypos+x*3+1] = 0;
	scaledpic[ypos+x*3+2] = 0;
      }
    }
  }

  /*
  for (y = 0; y < 200; y++) {
    ypos = y*320*3;
    for (x = 0; x < 320; x++) {
      scaledpic[ypos+x*3  ] = ((x * 255)/319) * ((y/16) & 1) / 1;
      scaledpic[ypos+x*3+1] = ((x * 255)/319) * ((y/16) & 2) / 2;
      scaledpic[ypos+x*3+2] = ((x * 255)/319) * ((y/16) & 4) / 4;
    }
  }
  */

  /* free the jpeg */
  free(rawimage);

  /* reduce with fs dither */
  reduce(scaledpic, pic);

  if (verbose) fprintf(stderr, "reduction done\n");

  /* it'd help if we did this before padding with black... */
  bkgcol = findbkg(pic);

  if (verbose) fprintf(stderr, "got bkg %d\n", bkgcol);

  /* here we do the actual conversion */
  {
    int x, y, z;
    UINT8 block[64];
    /* UINT8 dest[64]; */
    mcichar mc;
    unsigned int m;
    int gfxfile;
    UINT8 mem[0xd000];
    UINT8 colram[1000];
    UINT8 charram0[1000];
    UINT8 charram1[1000];
    UINT8 bitmap0[8000];
    UINT8 bitmap1[8000];

    /* init to background colour */
    lastcol0 = lastcol1 = lastcol2 = bkgcol;

    for (y = 0; y < 25; y++) {
      for (x = 0; x < 40; x++) {
	for (z = 0; z < 8; z++) {
	  /* read 8x8 pixels from the bitmap into a buffer */
	  block[z*8  ] = pic[(y*8+z)*320+x*8  ];
	  block[z*8+1] = pic[(y*8+z)*320+x*8+1];
	  block[z*8+2] = pic[(y*8+z)*320+x*8+2];
	  block[z*8+3] = pic[(y*8+z)*320+x*8+3];
	  block[z*8+4] = pic[(y*8+z)*320+x*8+4];
	  block[z*8+5] = pic[(y*8+z)*320+x*8+5];
	  block[z*8+6] = pic[(y*8+z)*320+x*8+6];
	  block[z*8+7] = pic[(y*8+z)*320+x*8+7];
 	}
	/* convert block, store converted results in mc and a copy
	   suitable for displaying in dest */
	convertmcichar(block, &mc);
	{
	  int a,i;

	  a = y*40 + x;
	  colram[a] = mc.colram;
	  charram0[a] = mc.charram0;
	  charram1[a] = mc.charram1;
	  for (i = 0; i < 8; i++) {
	    bitmap0[a*8 + i] = mc.bitmap0[i];
	    bitmap1[a*8 + i] = mc.bitmap1[i];
	  }
	}
      }
    }

    if (verbose) fprintf(stderr, "creating image\n");

    /* clear mem */
    for (m = 0; m < sizeof(mem); m++) {
      mem[m] = 0;
    }
    /* copy code */
    for (m = 0; m < sizeof(mcicode); m++) {
      mem[A_CODE + m] = mcicode[m];
    }
    /* copy chars and colram */
    for (m = 0; m < 1000; m++) {
      mem[A_COLRAM   + m] = colram[m];
      mem[A_CHARRAM0 + m] = charram0[m];
      mem[A_CHARRAM1 + m] = charram1[m];
    }
    /* copy bitmaps */
    for (m = 0; m < 8000; m++) {
      mem[A_BITMAP0 + m] = bitmap0[m];
      mem[A_BITMAP1 + m] = bitmap1[m];
    }
    /* background colour */
    mem[A_COLRAM + 1001] = bkgcol;
    /* start address */
    mem[A_START - 2] = A_START & 0xff;
    mem[A_START - 1] = A_START / 256;

    if (verbose) fprintf(stderr, "writing to stdout\n");

    /* open output files */
    if (fwrite(mem + A_START - 2, 1, A_END - A_START + 2, stdout) != A_END - A_START + 2) {
      if (!quiet) fprintf(stderr, "Write error!\n");
    }
  }

  if (verbose) fprintf(stderr, "done\n");

  /* done */
  //printf("OK\r\n");
  return(0);
}
